# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# TODO: PYT-1213 this import will eventually be deprecated
from contrast.falcon.middleware import FalconMiddleware  # pylint: disable=unused-import
